(* $Id: godi_query.ml 530 2005-04-04 11:58:34Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Godi_types
open Godi_exec
open Godi_versions
open Godi_plist
open Godi_file
open Printf


module StringSet = Set.Make(String) ;;

(*
let location loc =
  "File " ^ loc.Godi_make.loc_filename ^ ", line " ^
  string_of_int loc.Godi_make.loc_linenumber
;;


let error_msg exc =
  match exc with
      Failure s -> s
    | Command_failure(cmd,args,code) ->
	( match code with
	      WEXITED n ->
		sprintf "Command fails with code %d: %s" n cmd
	    | WSIGNALED n ->
		sprintf "Command terminated because of signal %d: %s" n cmd
	    | WSTOPPED _ ->
		sprintf "Command stopped: %s" cmd
	)
    | Command_not_found cmd ->
	sprintf "Command not found: %s" cmd
    | Command_discontinued ->
	"Command aborted"
    | General_error ->
	"Error"
    | Godi_make.Parse_error(loc, msg) ->
	sprintf "%s [Makefile parsing]: %s" (location loc) msg
    | Godi_make.Eval_error(loc, msg) ->
	sprintf "%s [Makefile evaluation]: %s" (location loc) msg
    | Godi_make.File_not_found file ->
	sprintf "File not found: %s" file
    | Godi_match.Bad_glob_expr s ->
	sprintf "Bad glob expression: %s" s
    | Godi_match.Bad_regular_expr s ->
	sprintf "Bad regular expression: %s" s
    | Godi_match.Unsupported_expr s ->
	sprintf "Unsupported glob/regular expression: %s" s
    | Godi_match.Bad_subst_expr ->
	"Bad substitution"
    | Godi_match.No_percent_symbol ->
	"Missing '%' as substitution identifier"
    | _ ->
	Printexc.to_string exc
;;
*)



let to_prot protocol_option msg =
  match protocol_option with
      None -> ()
    | Some p -> collect p msg
;;


let tws_re = Str.regexp "[ \t\r\n]*$";;

let strip_trailing_ws s =
  Str.global_replace tws_re "" s
;;


let lazy_etc_dir =
  lazy (
    let b = Buffer.create 100 in
    exec_to_user_error
      ~pref_user_error:true
      (call_cmd ~stdout:(`Buffer b)) ("godi_confdir", [| |]);
    strip_trailing_ws(Buffer.contents b)
  )
;;


let lazy_godi_conf =
  lazy (
    let etc_dir = Lazy.force lazy_etc_dir in
    Filename.concat etc_dir "godi.conf"
  )
;;


let lazy_localbase =
  lazy (
    let godi_conf = Lazy.force lazy_godi_conf in
    let b = Buffer.create 100 in
    exec_to_user_error
      ~pref_user_error:true
      (call_cmd
	 ~stdout:(`Buffer b))
      ("godi_make",
       [| "-f"; godi_conf; "-V"; "LOCALBASE"; "all"|]);
    strip_trailing_ws(Buffer.contents b)
  )
;;


let lazy_pkgdbdir =
  lazy (
    let godi_conf = Lazy.force lazy_godi_conf in
    let b = Buffer.create 100 in
    exec_to_user_error
      ~pref_user_error:true
      (call_cmd
	 ~stdout:(`Buffer b))
      ("godi_make",
       [| "-f"; godi_conf; "-V"; "GODI_DBDIR"; "all"|]);
    strip_trailing_ws(Buffer.contents b)
  )
;;


let godi_conf() = Lazy.force lazy_godi_conf
;;


let localbase() = Lazy.force lazy_localbase
;;


let pkgdbdir() = Lazy.force lazy_pkgdbdir
;;


let buildbase() = Filename.concat (localbase()) "build" ;;


let get_tools_makefile() =
  let cmd_name = Sys.argv.(0) in
  let base_cmd_name = Filename.basename cmd_name in
  if base_cmd_name = "godi_console_dev" then (
    let bin_dir = Filename.dirname cmd_name in
    let f = Filename.concat bin_dir "../godi-pkg/godi_tools.mk" in
    if f.[0] = '/' then
      f
    else
      Filename.concat (Sys.getcwd()) f
  ) 
  else (
    let bbase = buildbase() in
    let dir_mk = Filename.concat bbase "mk" in
    Filename.concat dir_mk "godi_tools.mk"
  )
;;


let lazy_pkgtools_version =
  lazy (
    let b = Buffer.create 100 in
    let file = get_tools_makefile() in
    exec_to_user_error
      ~pref_user_error:true
      (call_cmd
	 ~stdout:(`Buffer b))
      ("godi_make",
       [| "-f"; file; "print-pkgtools-version"|]);
    strip_trailing_ws(Buffer.contents b)
  )
;;


let pkgtools_version() = Lazy.force lazy_pkgtools_version ;;



let read_file ?(pref_user_error=false) filename =
  (* Return the contents of [filename] *)
  let f = 
    try
      open_in_bin filename 
    with
	Sys_error msg when pref_user_error ->
	  user_error("System error: " ^ msg)
  in
  try
    let n = in_channel_length f in
    let s = String.create n in
    really_input f s 0 n;
    close_in f;
    s
  with
      error -> close_in f; raise error
;;


let read_optional_file ?pref_user_error filename =
  (* Returns the contents of [filename], if existing, and "" otherwise *)
  if Sys.file_exists filename then
    read_file ?pref_user_error filename
  else
    ""
;;


(***********************************************************************)
(* Source packages                                                     *)
(***********************************************************************)

let lines_re = Str.regexp "\n";;

let sloppy_lines_re = Str.regexp "[ \t]*\n[ \t]*";;

let words_re = Str.regexp "[ \t]+";;

let dep_re = Str.regexp "\\([^:]*\\):";;



let parse_dep_string s =
  if Str.string_match dep_re s 0 then (
    let cond_str = Str.matched_group 1 s in
    parse_dependency cond_str
  )
  else
    parse_dependency s
;;


let parse_dep_strings s =
  let words = Str.split words_re s in
  List.map parse_dep_string words
;;


let classic_query_makefile ?pref_user_error ?protocol ?godi_conf ?(makefile="Makefile") dir variables =
  (* TODO: Write a quicker version that does not call "make" to get
   * the values of variables. A better method is to scan the Makefile,
   * and to interpret the definitions ourselves, and to fall back to
   * "make" only if we find conditional or computed definitions.
   *)
  let env =
    match godi_conf with
	None -> []
      | Some cf -> [ "GODI_CONF", cf ] in
  let b = Buffer.create 1000 in
  exec_to_user_error
    ?pref_user_error
    (call_cmd
       ~cd:dir
       ~env
       ?stderr:protocol
       ~stdout:(`Buffer b))
    ("godi_make",
     [| "-f"; makefile;
        "show-vars";
	"VARNAMES=" ^ String.concat " " variables |]);
  let s = Buffer.contents b in
  assert(s <> "" && s.[ String.length s - 1 ] = '\n');
  (* Remove the trailing \n: *)
  let s' = String.sub s 0 (String.length s - 1) in
  (* Split into lines: *)
  let l = Str.split_delim lines_re s' in
  assert(List.length l = List.length variables);
  l
;;


let pcache = Godi_make.create_parser_cache();;

let modern_query_makefile ?pref_user_error ?protocol ?godi_conf:gcarg ?(makefile="Makefile") dir variables =

  let print_error msg =
    match protocol with
	None ->
	  prerr_string msg;
	  flush stderr
      | Some p ->
	  Godi_exec.collect p msg
  in

  let gc =
    match gcarg with
	None    -> godi_conf()
      | Some cf -> cf
  in

  let env =
    Array.concat [Unix.environment(); [| "GODI_CONF=" ^  gc |] ] in
  (* Latter variables override previous instances ! *)

  (* print_error ("Doing " ^ dir ^ "\n"); *)

  (* Notes:
   * - There is always GODI_CONF in the environment
   * - PKGTOOLS_VERSION is also set
   * The result is that certain time-consuming actions need not to
   * be done when the Makefile is evaluated (running further external
   * programs)
   *)

  let ctx = Godi_make.create_context ~env ~curdir:dir ~objdir:dir
	      ~override:[ "PKGTOOLS_VERSION", pkgtools_version();
			]
	      ~make_cmd:"godi_make"
	      [] 
	      ?stderr:protocol
  in
  Godi_make.fast_parse_eval ?pref_user_error pcache makefile ctx;
  List.map
    (fun varname ->
       let loc = { Godi_make.loc_filename = "<Expansion of " ^ varname ^ ">";
		   Godi_make.loc_linenumber = 0 } in
       Godi_make.expand_var ?pref_user_error ctx loc varname
    )
    variables
;;

let query_makefile =
  (* classic_query_makefile *)
  modern_query_makefile ;;


let read_confopts dir =
  let fn = Filename.concat dir "CONFOPTS" in
  if Sys.file_exists fn then
    let c = read_file fn in
    List.filter (fun line -> line <> "") (Str.split sloppy_lines_re c)
  else
    []
;;


module VId = struct
  type t = var_id
  let compare = (compare : t -> t -> int)
end;;

module VIdMap = Map.Make(VId);;


class source_package_impl ?pref_user_error ?protocol (dir:string) : source_package =
  let confopts = read_confopts dir in
  let values =
    query_makefile
      ?pref_user_error
      ?protocol
      dir (["PKGNAME"; "COMMENT"; "MAINTAINER"; "DEPENDS"; "BUILD_DEPENDS"; "CONFLICTS";
	    "SELF_UPDATE_PKG";
	   ] @ confopts) in

  let (fullname,comment,maintainer,depends,build_depends,conflicts,self_upd_pkg,
       confopts_vals) =
    match values with
	fullname :: comment :: maintainer :: depends :: build_depends :: conflicts :: 
	self_upd_pkg ::
	confopts_vals ->
	    (fullname, comment, maintainer, depends, build_depends, conflicts, 
	     self_upd_pkg,
	     confopts_vals)
      | _ ->
	  assert false
  in

  (* CHECK: Maybe we should catch errors in the following expressions: *)
  let (name, version, pkgrevision) = split_name_version fullname in
  let init_dependencies = parse_dep_strings depends in
  let init_build_dependencies = parse_dep_strings build_depends in
  let init_conflicts = parse_dep_strings conflicts in

object(self)
  val mutable user_status = `Keep
  val mutable user_options = None

  val dependencies = init_dependencies
  val build_dependencies = init_build_dependencies
  val conflicts = init_conflicts

  val mutable config_dialog_shown = false

  val var_id = new_var_id()
  val mutable variants = VIdMap.empty

  initializer
    variants <- VIdMap.add var_id self variants

  method name = name
  method version = version
  method pkgrevision = pkgrevision
  method comment = comment
  method godi_maintainer = maintainer
  method description =
    read_optional_file ?pref_user_error (Filename.concat dir "DESCR")

  method dependencies = dependencies
  method build_dependencies = build_dependencies
  method conflicts = conflicts
  method var_id = var_id

  method self_update_pkg = self_upd_pkg
  method variant vid =
    try
      VIdMap.find vid variants
    with
	Not_found ->
	  failwith "class source_package: No such variant"

  method user_status = user_status
  method set_user_status s =
    match s with
	`Keep | `Build | `Build_dep _ ->
	  user_status <- s
      | _ ->
	  failwith "Bad user_status"

  method source_directory = dir

  method build_message =
    read_optional_file ?pref_user_error (Filename.concat dir "BUILDMSG")

  method build_options =
    confopts

  method values_of_build_options =
    List.combine confopts confopts_vals

  method values_of_user_build_options =
    match user_options with
	Some opts -> opts
      | None -> self#values_of_build_options

  method set_user_build_options opts =
    let names = List.map fst opts in
    if names <> self#build_options then
      invalid_arg "set_user_build_options";

    user_options <- Some opts

  method config_dialog_shown = config_dialog_shown

  method set_config_dialog_shown b = config_dialog_shown <- b

  method rescan_deps ?protocol godi_conf_file vid =
    let values =
      query_makefile
	?pref_user_error
	?protocol
	~godi_conf:godi_conf_file
	dir [ "DEPENDS"; "BUILD_DEPENDS"] in

    let v_deps, v_bld_deps =
      match values with
	  [ v_deps; v_bld_deps ] -> (v_deps, v_bld_deps)
	| _ -> assert false in

    let new_pkg =
      {< dependencies = parse_dep_strings v_deps;
	 build_dependencies = parse_dep_strings v_bld_deps;
	 var_id = vid
      >} in

    variants <- VIdMap.add vid new_pkg variants;
    ()

end
;;


exception Next of string list
;;

let list_subdirs dirname =
  (* Return the list of all subdirectories of [dirname]; the returned names
   * include [dirname].
   *)
  (* TODO: Base implementation on Godi_file *)
  let rec read d l =
    try
      let x = Unix.readdir d in
      if x = "." || x = ".." then raise(Next []);
      let n = Filename.concat dirname x in
      let stat = Unix.stat n in
      if stat.Unix.st_kind = Unix.S_DIR then
	raise(Next [n])
      else
	raise(Next [])
    with
	End_of_file ->
	  l
      | Next xl ->
	  read d (xl @ l)
      | Unix.Unix_error(Unix.ENOENT,_,_) ->  (* e.g. bad symlink *)
	  read d l
  in
  let d = Unix.opendir dirname in
  try
    let l = read d [] in
    Unix.closedir d;
    l
  with
      error -> Unix.closedir d; raise error
;;


let really_scan_source_packages ?protocol () =
  (* To get a list of all source packages, traverse the subdirectories
   * at LOCALBASE/build. All directories two levels below this location
   * that have a "Makefile" and a "DESCR" file are considered as
   * source package.
   * This is a lousy definition!
   *)
  let build = Filename.concat (localbase()) "build" in
  (* TODO: Handle errors from list_subdirs: *)
  let level1 =
    list_subdirs build in
  let level2 =
    List.flatten
      (List.map (fun dirname -> list_subdirs dirname) level1) in
  let source_dirs =
    List.filter
      (fun dirname ->
	 Sys.file_exists(Filename.concat dirname "Makefile") &&
	 Sys.file_exists(Filename.concat dirname "DESCR"))
      level2 in


  List.flatten
    (List.map
       (fun dir ->
	  try
	    let src_pkg = new source_package_impl 
			    ~pref_user_error:true ?protocol dir in
	    [ `Source_pkg src_pkg ]
	  with
	      User_error msg ->
		to_prot protocol
		  (sprintf
		     "Scan of directory %s: %s\n" dir msg);
		[]
       )
       source_dirs)
;;


let source_packages = ref None;;


let scan_source_packages ?(rescan=false) ?protocol () =
  if rescan then source_packages := None;
  match !source_packages with
      None ->
	let pkg_list = really_scan_source_packages ?protocol () in
	source_packages := Some pkg_list;
	pkg_list
    | Some pkg_list ->
	pkg_list
;;


let forget_source_packages() =
  source_packages := None
;;


(***********************************************************************)
(* Installed packages                                                  *)
(***********************************************************************)

let name_re = Str.regexp "^@name[ \t]+\\([^ \t]+\\)[ \t]*$"
let pkgdep_re = Str.regexp "^@pkgdep[ \t]+\\([^ \t]+\\)[ \t]*$"
let blddep_re = Str.regexp "^@blddep[ \t]+\\([^ \t]+\\)[ \t]*$"


let from_tgz component tgz =
  try
    let b = Buffer.create 100 in
    Godi_exec.call_cmd
      ~stdout:(`Buffer b)
      ~stderr:(`Forward (fun _ -> ()))  (* discard *)
      ( "godi_tar", [| "-xzqOf"; tgz; component |] );
    Some(Buffer.contents b)
  with
      Command_not_found s ->
	user_error("Command not found: " ^ s)
    | Command_failure(_,_,_) ->
	None
;;


let parse_CONTENTS ?(pref_user_error=false) ?no_files dir =
  let filename = Filename.concat dir "+CONTENTS" in
  try
    parse_inst_plist ?no_files filename
  with
      Sys_error msg when pref_user_error ->
	user_error("System error: " ^ msg)
    | Plist_error(file, line) when pref_user_error ->
	user_error("Bad line in file `" ^ file ^ "': " ^ line)
;;


let parse_CONTENTS_from_tgz ?(pref_user_error=false) ?no_files tgz =
  (* This version is used by [binary_packege_impl] *)
  let plist_str = 
    match from_tgz "+CONTENTS" tgz with
	Some s -> s
      | None ->
	  user_error("This file is not a binary GODI package: " ^ tgz)
  in
  try
    parse_inst_plist_string ?no_files plist_str
  with
    | Plist_error(file, line) when pref_user_error ->
	user_error("Bad line in `+CONTENTS' in `" ^ tgz ^ "': " ^ line)
;;


let extract_name_from_plist ?(pref_user_error=false) plist =
  try
    match 
      List.find (function `Name _ -> true | _ -> false) plist
    with
	`Name n -> n
      | _ -> assert false
    with
	Not_found when pref_user_error ->
	  user_error("Missing @name directive in +CONTENTS")
;;


let extract_pkgdep_from_plist plist =
  List.flatten
    (List.map
       (function `Pkgdep d -> [d] | _ -> []) 
       plist
    )
;;


let extract_blddep_from_plist plist =
  List.flatten
    (List.map
       (function `Blddep d -> [d] | _ -> []) 
       plist
    )
;;


let extract_pkgcfl_from_plist plist =
  List.flatten
    (List.map
       (function `Pkgcfl d -> [d] | _ -> []) 
       plist
    )
;;


type inst_set_status =
    [ `Reload
    | `Reconcile of package list
    | `Up_to_date of package list
    ]
;;


let installed_packages = ref (`Reload : inst_set_status);;

let get_installed_packages_fn = ref (fun () -> assert false) ;;
  (* "forward" declaration *)


class installed_package_impl ?pref_user_error (dir:string) : installed_package =
  let mtime = file_mtime dir in
  let plist = parse_CONTENTS ?pref_user_error ~no_files:true dir in
  let fullname = extract_name_from_plist ?pref_user_error plist in
  let dependencies = extract_pkgdep_from_plist plist in
  let build_dependencies = extract_blddep_from_plist plist in
  let conflicts = extract_pkgcfl_from_plist plist in

  let comment = read_file ?pref_user_error (Filename.concat dir "+COMMENT") in
  let size =
    try Int64.of_string 
          (strip_trailing_ws (read_file (Filename.concat dir "+SIZE_PKG")))
    with
	Sys_error _ -> 0L
      | Failure _ ->   0L in

  let size_all =
    try Int64.of_string 
          (strip_trailing_ws (read_file (Filename.concat dir "+SIZE_ALL")))
    with
	Sys_error _ -> 0L
      | Failure _ ->   0L in

  let preserve = Sys.file_exists (Filename.concat dir "+PRESERVE") in

  let [ req_script_path, req_script;
	inst_script_path, inst_script;
	deinst_script_path, deinst_script ] =
    List.map
      (fun b ->
	 let p = Filename.concat dir b in
	 if Sys.file_exists p then 
	   (Some p, Some (read_file ?pref_user_error p) )
	 else 
	   (None, None)
      )
      [ "+REQUIRE"; "+INSTALL"; "+DEINSTALL" ] in

  let (name, version, pkgrevision) = split_name_version fullname in

object(self:'self)
  val mutable user_status = `Keep

  method name = name
  method version = version
  method pkgrevision = pkgrevision
  method comment = comment
  method description = read_file ?pref_user_error (Filename.concat dir "+DESC")
  method godi_maintainer = "(Info n/a for non-source packages)"

  method db_directory = dir
  method preserve = preserve
  method inst_plist = parse_CONTENTS ?pref_user_error dir
  method size = size
  method size_all = size_all
  method requirements_script_path = req_script_path
  method requirements_script = req_script
  method install_script_path = inst_script_path
  method install_script = inst_script
  method deinstall_script_path = deinst_script_path
  method deinstall_script = deinst_script

  method dependencies = dependencies
  method build_dependencies = build_dependencies
  method conflicts = conflicts

  method db_entry_last_modified = mtime

  method user_status = user_status
  method set_user_status s =
    match s with
	`Keep | `Remove | `Remove_dep _ ->
	  user_status <- s
      | _ ->
	  failwith "Bad user_status"

  method var_id = new_var_id()

  method variant (v:var_id) : 'self =
    failwith "class installed_package: No such variant"

  method update_successor_file ?(filter = fun _ -> true) () =
    (* Find out all packages that depend on this package: *)
    let all_installed = !get_installed_packages_fn () in
    let successors =
      List.filter
	(fun ipkg ->
	   (List.exists (fun (n,_) -> n = name) (pkg_dependencies ipkg)) &&
	   filter (pkg_name ipkg)
	)
	all_installed in
    let f = open_out (Filename.concat dir "+REQUIRED_BY") in
    ( try
	List.iter
	  (fun succ ->
	     fprintf f "%s\n" (pkg_fullname succ)
	  )
	  successors;
	close_out f
      with
	  error -> close_out f; raise error
    )

end
;;


let update_installed_packages ?protocol ?(verbose = false) () =
  (* To get a list of all source packages, traverse the subdirectories
   * at PKG_DBDIR. All subdirectories that have a "+CONTENTS" file
   * are considered as installed package.
   *)
  let message s =
    if verbose then (
      match protocol with
	  Some p -> collect p s
	| None -> ()
    ) in

  let dbdir = pkgdbdir() in
  (* TODO: Handle errors from list_subdirs: *)
  let level1 =
    list_subdirs dbdir in
  let inst_dirs =
    List.filter
      (fun dirname ->
	 Sys.file_exists(Filename.concat dirname "+CONTENTS"))
      level1 in

  let load_single_pkg dir =
    try
      let inst_pkg = new installed_package_impl 
		       ~pref_user_error:true dir in
      [ `Installed_pkg inst_pkg ]
    with
	User_error msg ->
	  to_prot protocol
	  (sprintf
	     "Scan of directory %s: %s\n" dir msg);
	  []
  in

  let load dirs =
    List.flatten (List.map load_single_pkg dirs) in

  let update pl =
    (* Iterate over pl, and check whether the package exists and is
     * up-to-date. Returns a new list.
     *)
    List.flatten
      (List.map
	 (fun pkg ->
	    let inst_pkg = installed_pkg pkg in
	    if Sys.file_exists inst_pkg#db_directory then (
	      let mtime = file_mtime inst_pkg#db_directory in
	      if mtime <> inst_pkg#db_entry_last_modified then (
		load_single_pkg inst_pkg#db_directory  
                  (* Package was modified *)
	      )
	      else
		[ pkg ]   (* Package not modified *)
	    )
	    else
	      []   (* Package was deleted *)
	 )
	 pl
      )
  in

  let reconcile pl =
    (* First check the existing packages: *)
    let pl' = update pl in
    (* Now check whether there are new packages. Create a set of all known
     * directories, compare with inst_dirs
     *)
    let set =
      List.fold_left
	(fun acc pkg -> 
	   let dir = (installed_pkg pkg) # db_directory in
	   StringSet.add dir acc)
	StringSet.empty
	pl' in
    let new_dirs =
      List.filter
	(fun dir -> not (StringSet.mem dir set)) inst_dirs in
    pl' @ (load new_dirs)
  in

  match !installed_packages with
      `Reload ->
	message "Scanning system for installed packages...\n";
	let pl = load inst_dirs in
	installed_packages := `Up_to_date pl
    | `Reconcile old ->
	message "Scanning installed packages for changes...\n";
	let pl = reconcile old in
	installed_packages := `Up_to_date pl
    | `Up_to_date current ->
	()
;;


let get_installed_packages ?protocol ?verbose () =
  update_installed_packages ?protocol ?verbose ();
  match !installed_packages with
      `Up_to_date current ->
	current
    | _ ->
	assert false
;;


get_installed_packages_fn := get_installed_packages ;;


let get_this_installed_package ?protocol ?verbose name =
  let all_pkgs = get_installed_packages ?protocol ?verbose () in
  List.find
    (fun pkg -> (pkg_name pkg) = name)
    all_pkgs
;;


let notify_about_change_of_installed_packages =
  function
      `Reload ->
	installed_packages := `Reload
    | `Reconcile ->
	let pl = get_installed_packages() in
	installed_packages := `Reconcile pl
;;

(***********************************************************************)
(* Binary packages                                                     *)
(***********************************************************************)

class binary_package_impl ?pref_user_error (tgz:string) : binary_package =
  let tgz = 
    if Filename.is_relative tgz then
      Filename.concat (Sys.getcwd()) tgz 
    else
      tgz in
  let plist = parse_CONTENTS_from_tgz ?pref_user_error ~no_files:true tgz in
  let fullname = extract_name_from_plist ?pref_user_error plist in
  let dependencies = extract_pkgdep_from_plist plist in
  let build_dependencies = extract_blddep_from_plist plist in
  let conflicts = extract_pkgcfl_from_plist plist in

  let comment = 
    match from_tgz "+COMMENT" tgz with
	Some c -> c
      | None   -> "" in
(*
  let size =
    try Int64.of_string 
          (strip_trailing_ws (read_file (Filename.concat dir "+SIZE_PKG")))
    with
	Sys_error _ -> 0L
      | Failure _ ->   0L in

  let size_all =
    try Int64.of_string 
          (strip_trailing_ws (read_file (Filename.concat dir "+SIZE_ALL")))
    with
	Sys_error _ -> 0L
      | Failure _ ->   0L in
  let preserve = Sys.file_exists (Filename.concat dir "+PRESERVE") in

  let [ req_script_path, req_script;
	inst_script_path, inst_script;
	deinst_script_path, deinst_script ] =
    List.map
      (fun b ->
	 let p = Filename.concat dir b in
	 if Sys.file_exists p then 
	   (Some p, Some (read_file ?pref_user_error p) )
	 else 
	   (None, None)
      )
      [ "+REQUIRE"; "+INSTALL"; "+DEINSTALL" ] in
*)

  let (name, version, pkgrevision) = split_name_version fullname in

object(self:'self)
  val mutable user_status = `Keep

  method name = name
  method version = version
  method pkgrevision = pkgrevision
  method comment = comment
  method description = assert false
    (* read_file ?pref_user_error (Filename.concat dir "+DESC") *)
  method godi_maintainer = "(Info n/a for non-source packages)"

  method binary_archive = tgz
  method require_binary_archive ?protocol ?interactivity () = ()
  method release_binary_archive() = ()

  method dependencies = dependencies
  method build_dependencies = build_dependencies
  method conflicts = conflicts

  method user_status = user_status
  method set_user_status s =
    match s with
	`Keep | `Add | `Add_dep _ ->
	  user_status <- s
      | _ ->
	  failwith "Bad user_status"

  method var_id = new_var_id()

  method variant (v:var_id) : 'self =
    failwith "class binary_package: No such variant"

end
;;


let get_binary_package tgz =
  `Binary_pkg(new binary_package_impl ~pref_user_error:true tgz)
;;


(***********************************************************************)


let matches_dep pkg (name,cond) =
  let p = any_pkg pkg in
  p#name = name && version_matches_condition p#version cond
;;
